# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Placeholder for old entity_schema module."""

from .._version import VERSION

__version__ = VERSION
__author__ = "Ian Hellen"

# pylint: disable=wildcard-import, unused-wildcard-import
from ..datamodel.entities import *  # noqa: F403, F401
