# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTIC Anomalous Sequence Modelling Utilities."""

from ...._version import VERSION

__version__ = VERSION
